(* version tree implementation of functional arrays
   extended by a cache to speed up single-threaded use *)

structure FunArray:FUN_ARRAY = 
struct

  (* An array is either the (immutable) original array (Root of ...)
     or a modified array (Node of ...). A modified array consists
     of the index to be modified (int), the new value for this 
     index ('a), and the old array ('a array).
     
     So far this is the original structure of v-arrays. Search goes from
     a modification node towards the original array until a node with
     the searched index is found, or the original array is reached.
     
     A c-array stores an additional array used as a cache for the
     "last" updates on the functional array. Assume the updates to 
     an array are ordered left-to-right, then the array cache contains 
     the values of the array represented by the minimum (leftmost) node
     in the array tree. Initially, the cache contains the initial array
     values. Now searches in the leftmost array can be served by the
     cache. In particular, this means that single-threaded arrays (which
     are characterized by being a path in the v-array-representation)
     can be updated and searched in constant time. Since any but the first
     update of an array cannot make any use of the cache, we use two
     constructors: 
       Cache: for nodes on the left spine of the array version tree.
              These may use the cache
       Node:  for all other nodes which cannot use the cache.
     This saves a little space.
     
     How do we know that we can use the cache? When has the cache to
     be updated? We keep and propagate a boolean flag which is set to 
     true only for the leftmost node.
  *)     
     
  datatype 'a array = 
      Root  of 'a Array.array
    | Node  of int * 'a * 'a array
    | Cache of int * 'a * 'a array * bool ref * 'a Array.array
 
  fun array (n,x) = Cache (0,x,Root (Array.array (n,x)),
                           ref true,Array.array (n,x))
  
  fun search (Cache (_,_,Root a,_,_),i) = Array.sub (a,i)
   |  search (Cache (j,x,  tree,_,_),i) =
         if i=j then x else search (tree,i)
   |  search (Node  (_,_,Root a),i) = Array.sub (a,i)
   |  search (Node  (j,x,  tree),i) =
         if i=j then x else search (tree,i)
 
  and sub (tree as Cache (_,_,_,ref cache,c),i) =
          if cache then Array.sub (c,i) else search (tree,i)
   |  sub (tree,i) = search (tree,i)

  fun size (Root a)            = Array.length a
   |  size (Node (_,_,a))      = size a
   |  size (Cache (_,_,_,_,a)) = Array.length a
   
  fun update (a as Cache (_,_,_,cache,c),i,x) =
      if !cache then 
         (cache := false; Array.update (c,i,x); Cache (i,x,a,ref true,c))
      else 
         Node (i,x,a)
   |  update (a,i,x) = Node (i,x,a)
   
  fun fromList l = Cache (0,hd l,Root (Array.fromList l),  
                          ref true,Array.fromList l)

  fun fromImpArray a = 
      let val b = Array.array (Array.length a,Array.sub (a,0))
       in (Array.copy {src=a,si=0,len=NONE,dst=b,di=0};
           Cache (0,Array.sub (a,0),Root a,ref true,b))
      end

  exception NotImplemented
  fun toImpArray (Cache (_,_,Root a,_,_)) = 
      let val b = Array.array (Array.length a,Array.sub (a,0))
       in (Array.copy {src=a,si=0,len=NONE,dst=b,di=0}; b) end
   |  toImpArray _ =
      (print "toImpArray is currently just defined for \
              \ unchanged functional arrays";
       raise NotImplemented)          
end
;
